@@ -43,16 +43,16 @@
 // maximum to allocate on the stack, arbitrary limit
 #define SMALL_XML	65536
 #define MAX_STACK	100
-//#define BATCH_SIZE	(4096 / sizeof(struct _Leaf) - 1)
+//#define BATCH_SIZE	(4096 / sizeof(struct _leaf) - 1)
 #define BATCH_SIZE	100
 
-typedef struct _Batch {
-    struct _Batch	*next;
+typedef struct _batch {
+    struct _batch	*next;
     int			next_avail;
-    struct _Leaf	leaves[BATCH_SIZE];
+    struct _leaf	leaves[BATCH_SIZE];
 } *Batch;
 
-typedef struct _Doc {
+typedef struct _doc {
     Leaf		data;
     Leaf		*where;	     // points to current location
     Leaf		where_path[MAX_STACK]; // points to head of path
@@ -60,10 +60,10 @@ typedef struct _Doc {
     unsigned long	size;	     // number of leaves/branches in the doc
     VALUE		self;
     Batch		batches;
-    struct _Batch	batch0;
+    struct _batch	batch0;
 } *Doc;
 
-typedef struct _ParseInfo {
+typedef struct _parseInfo {
     char	*str;		/* buffer being read from */
     char	*s;		/* current position in buffer */
     Doc		doc;
@@ -92,7 +92,7 @@ static char*	read_quoted_value(ParseInfo pi);
 static void	skip_comment(ParseInfo pi);
 
 static VALUE	protect_open_proc(VALUE x);
-static VALUE	parse_json(VALUE clas, char *json, int given, int allocated);
+static VALUE	parse_json(VALUE clas, char *json, bool given, bool allocated);
 static void	each_leaf(Doc doc, VALUE self);
 static int	move_step(Doc doc, const char *path, int loc);
 static Leaf	get_doc_leaf(Doc doc, const char *path);
@@ -115,10 +115,10 @@ static VALUE	doc_each_value(int argc, VALUE *argv, VALUE self);
 static VALUE	doc_dump(int argc, VALUE *argv, VALUE self);
 static VALUE	doc_size(VALUE self);
 
-VALUE	oj_doc_class = 0;
+VALUE	oj_doc_class = Qundef;
 
 // This is only for CentOS 5.4 with Ruby 1.9.3-p0.
-#ifdef NEEDS_STPCPY
+#ifndef HAVE_STPCPY
 char *stpcpy(char *dest, const char *src) {
     size_t	cnt = strlen(src);
     
@@ -204,10 +204,10 @@ leaf_new(Doc doc, int type) {
     Leaf	leaf;
 
     if (0 == doc->batches || BATCH_SIZE == doc->batches->next_avail) {
-	Batch	b = ALLOC(struct _Batch);
+	Batch	b = ALLOC(struct _batch);
 
 	// Initializes all leaves with a NO_VAL value_type
-	memset(b, 0, sizeof(struct _Batch));
+	memset(b, 0, sizeof(struct _batch));
 	b->next = doc->batches;
 	doc->batches = b;
 	b->next_avail = 0;
@@ -358,7 +358,7 @@ leaf_float_value(Leaf leaf) {
 
 static VALUE
 leaf_array_value(Doc doc, Leaf leaf) {
-    VALUE	a = rb_ary_new();
+    volatile VALUE	a = rb_ary_new();
 
     if (0 != leaf->elements) {
 	Leaf	first = leaf->elements->next;
@@ -374,12 +374,12 @@ leaf_array_value(Doc doc, Leaf leaf) {
 
 static VALUE
 leaf_hash_value(Doc doc, Leaf leaf) {
-    VALUE	h = rb_hash_new();
+    volatile VALUE	h = rb_hash_new();
 
     if (0 != leaf->elements) {
-	Leaf	first = leaf->elements->next;
-	Leaf	e = first;
-	VALUE	key;
+	Leaf		first = leaf->elements->next;
+	Leaf		e = first;
+	volatile VALUE	key;
 
 	do {
 	    key = rb_str_new2(e->key);
@@ -734,7 +734,7 @@ read_quoted_value(ParseInfo pi) {
 // doc support functions
 inline static void
 doc_init(Doc doc) {
-    memset(doc, 0, sizeof(struct _Doc));
+    memset(doc, 0, sizeof(struct _doc));
     doc->where = doc->where_path;
     doc->self = Qundef;
     doc->batches = &doc->batch0;
@@ -778,17 +778,53 @@ free_doc_cb(void *x) {
     }
 }
 
+static void
+mark_leaf(Leaf leaf) {
+    switch (leaf->value_type) {
+    case COL_VAL:
+	if (NULL != leaf->elements) {
+	    Leaf	first = leaf->elements->next;
+	    Leaf	e = first;
+
+	    do {
+		mark_leaf(e);
+		e = e->next;
+	    } while (e != first);
+	}
+	break;
+    case RUBY_VAL:
+	rb_gc_mark(leaf->value);
+	break;
+
+    default:
+	break;
+    }
+}
+
+static void
+mark_doc(void *ptr) {
+    if (NULL != ptr) {
+	Doc	doc = (Doc)ptr;
+	
+	rb_gc_mark(doc->self);
+	mark_leaf(doc->data);
+    }
+}
+
 static VALUE
-parse_json(VALUE clas, char *json, int given, int allocated) {
-    struct _ParseInfo	pi;
-    VALUE		result = Qnil;
+parse_json(VALUE clas, char *json, bool given, bool allocated) {
+    struct _parseInfo	pi;
+    volatile VALUE	result = Qnil;
     Doc			doc;
     int			ex = 0;
+    volatile VALUE	self;
 
+    // TBD are both needed? is stack allocation ever needed?
+    
     if (given) {
-	doc = ALLOCA_N(struct _Doc, 1);
+	doc = ALLOCA_N(struct _doc, 1);
     } else {
-	doc = ALLOC_N(struct _Doc, 1);
+	doc = ALLOC(struct _doc);
     }
     /* skip UTF-8 BOM if present */
     if (0xEF == (uint8_t)*json && 0xBB == (uint8_t)json[1] && 0xBF == (uint8_t)json[2]) {
@@ -805,7 +841,7 @@ parse_json(VALUE clas, char *json, int given, int allocated) {
     {
 	struct rlimit	lim;
 
-	if (0 == getrlimit(RLIMIT_STACK, &lim)) {
+	if (0 == getrlimit(RLIMIT_STACK, &lim) && RLIM_INFINITY != lim.rlim_cur) {
 	    pi.stack_min = (void*)((char*)&lim - (lim.rlim_cur / 4 * 3)); // let 3/4ths of the stack be used only
 	} else {
 	    pi.stack_min = 0; // indicates not to check stack limit
@@ -813,22 +849,22 @@ parse_json(VALUE clas, char *json, int given, int allocated) {
     }
 #endif
     // last arg is free func void* func(void*)
-#if HAS_DATA_OBJECT_WRAP
-    doc->self = rb_data_object_wrap(clas, doc, 0, free_doc_cb);
+#ifdef HAVE_RB_DATA_OBJECT_WRAP
+    self = rb_data_object_wrap(clas, doc, mark_doc, free_doc_cb);
 #else
-    doc->self = rb_data_object_alloc(clas, doc, 0, free_doc_cb);
+    self = rb_data_object_alloc(clas, doc, mark_doc, free_doc_cb);
 #endif
-    rb_gc_register_address(&doc->self);
+    doc->self = self;
     doc->json = json;
     DATA_PTR(doc->self) = doc;
     result = rb_protect(protect_open_proc, (VALUE)&pi, &ex);
     if (given || 0 != ex) {
-	rb_gc_unregister_address(&doc->self);
-	DATA_PTR(doc->self) = 0;
+	DATA_PTR(doc->self) = NULL;
 	doc_free(pi.doc);
 	if (allocated && 0 != ex) { // will jump so caller will not free
 	    xfree(json);
 	}
+	rb_gc_enable();
     } else {
 	result = doc->self;
     }
@@ -1099,14 +1135,14 @@ each_value(Doc doc, Leaf leaf) {
 
 // doc functions
 
-/* call-seq: open(json) { |doc| ... } => Object
+/* @overload open(json) { |doc| ... } => Object
  *
  * Parses a JSON document String and then yields to the provided block if one
  * is given with an instance of the Oj::Doc as the single yield parameter. If
  * a block is not given then an Oj::Doc instance is returned and must be
  * closed with a call to the #close() method when no longer needed.
  *
- * @param [String] json JSON document string
+ *   @param [String] json JSON document string
  * @yieldparam [Oj::Doc] doc parsed JSON document
  * @yieldreturn [Object] returns the result of the yield as the result of the method call
  * @example
@@ -1118,36 +1154,42 @@ each_value(Doc doc, Leaf leaf) {
  */
 static VALUE
 doc_open(VALUE clas, VALUE str) {
-    char	*json;
-    size_t	len;
-    VALUE	obj;
-    int		given = rb_block_given_p();
-    int		allocate;
+    char		*json;
+    size_t		len;
+    volatile VALUE	obj;
+    int			given = rb_block_given_p();
+    int			allocate;
 
     Check_Type(str, T_STRING);
-    len = RSTRING_LEN(str) + 1;
+    len = (int)RSTRING_LEN(str) + 1;
     allocate = (SMALL_XML < len || !given);
     if (allocate) {
 	json = ALLOC_N(char, len);
     } else {
 	json = ALLOCA_N(char, len);
     }
+    // It should not be necessaary to stop GC but if it is not stopped and a
+    // large string is parsed that string is corrupted or freed during
+    // parsing. I'm not sure what is going on exactly but disabling GC avoids
+    // the issue.
+    rb_gc_disable();
     memcpy(json, StringValuePtr(str), len);
     obj = parse_json(clas, json, given, allocate);
+    rb_gc_enable();
     if (given && allocate) {
 	xfree(json);
     }
     return obj;
 }
 
-/* call-seq: open_file(filename) { |doc| ... } => Object
+/* @overload open_file(filename) { |doc| ... } => Object
  *
  * Parses a JSON document from a file and then yields to the provided block if
  * one is given with an instance of the Oj::Doc as the single yield
  * parameter. If a block is not given then an Oj::Doc instance is returned and
  * must be closed with a call to the #close() method when no longer needed.
  *
- * @param [String] filename name of file that contains a JSON document
+ *   @param [String] filename name of file that contains a JSON document
  * @yieldparam [Oj::Doc] doc parsed JSON document
  * @yieldreturn [Object] returns the result of the yield as the result of the method call
  * @example
@@ -1160,13 +1202,13 @@ doc_open(VALUE clas, VALUE str) {
  */
 static VALUE
 doc_open_file(VALUE clas, VALUE filename) {
-    char	*path;
-    char	*json;
-    FILE	*f;
-    size_t	len;
-    VALUE	obj;
-    int		given = rb_block_given_p();
-    int		allocate;
+    char		*path;
+    char		*json;
+    FILE		*f;
+    size_t		len;
+    volatile VALUE	obj;
+    int			given = rb_block_given_p();
+    int			allocate;
 
     Check_Type(filename, T_STRING);
     path = StringValuePtr(filename);
@@ -1189,7 +1231,9 @@ doc_open_file(VALUE clas, VALUE filename) {
     }
     fclose(f);
     json[len] = '\0';
+    rb_gc_disable();
     obj = parse_json(clas, json, given, allocate);
+    rb_gc_enable();
     if (given && allocate) {
 	xfree(json);
     }
@@ -1223,7 +1267,7 @@ append_key(char *p, const char *key) {
  * @see Oj::Doc.open
  */
 
-/* call-seq: where?() => String
+/* @overload where?() => String
  *
  * Returns a String that describes the absolute path to the current location
  * in the JSON document.
@@ -1261,11 +1305,12 @@ doc_where(VALUE self) {
 	    *p++ = '/';
 	}
 	*--p = '\0';
+
 	return rb_str_new(path, p - path);
     }
 }
 
-/* call-seq: local_key() => String, Fixnum, nil
+/* @overload local_key() => String, Fixnum, nil
  *
  * Returns the final key to the current location.
  * @example
@@ -1275,9 +1320,9 @@ doc_where(VALUE self) {
  */
 static VALUE
 doc_local_key(VALUE self) {
-    Doc		doc = self_doc(self);
-    Leaf	leaf = *doc->where;
-    VALUE	key = Qnil;
+    Doc			doc = self_doc(self);
+    Leaf		leaf = *doc->where;
+    volatile VALUE	key = Qnil;
 
     if (T_HASH == leaf->parent_type) {
 	key = rb_str_new2(leaf->key);
@@ -1288,7 +1333,7 @@ doc_local_key(VALUE self) {
     return key;
 }
 
-/* call-seq: home() => nil
+/* @overload home() => nil
  *
  * Moves the document marker or location to the hoot or home position. The
  * same operation can be performed with a Oj::Doc.move('/').
@@ -1305,13 +1350,13 @@ doc_home(VALUE self) {
     return oj_slash_string;
 }
 
-/* call-seq: type(path=nil) => Class
+/* @overload type(path=nil) => Class
  *
  * Returns the Class of the data value at the location identified by the path
  * or the current location if the path is nil or not provided. This method
  * does not create the Ruby Object at the location specified so the overhead
  * is low.
- * @param [String] path path to the location to get the type of if provided
+ *   @param [String] path path to the location to get the type of if provided
  * @example
  *   Oj::Doc.open('[1,2]') { |doc| doc.type() }	     #=> Array
  *   Oj::Doc.open('[1,2]') { |doc| doc.type('/1') }  #=> Fixnum
@@ -1347,24 +1392,24 @@ doc_type(int argc, VALUE *argv, VALUE self) {
     return type;
 }
 
-/* call-seq: fetch(path=nil) => nil, true, false, Fixnum, Float, String, Array, Hash
+/* @overload fetch(path=nil) => nil, true, false, Fixnum, Float, String, Array, Hash
  *
  * Returns the value at the location identified by the path or the current
  * location if the path is nil or not provided. This method will create and
  * return an Array or Hash if that is the type of Object at the location
  * specified. This is more expensive than navigating to the leaves of the JSON
  * document.
- * @param [String] path path to the location to get the type of if provided
+ *   @param [String] path path to the location to get the type of if provided
  * @example
  *   Oj::Doc.open('[1,2]') { |doc| doc.fetch() }      #=> [1, 2]
  *   Oj::Doc.open('[1,2]') { |doc| doc.fetch('/1') }  #=> 1
  */
 static VALUE
 doc_fetch(int argc, VALUE *argv, VALUE self) {
-    Doc		doc;
-    Leaf	leaf;
-    VALUE	val = Qnil;
-    const char	*path = 0;
+    Doc			doc;
+    Leaf		leaf;
+    volatile VALUE	val = Qnil;
+    const char		*path = 0;
 
     doc = self_doc(self);
     if (1 <= argc) {
@@ -1380,12 +1425,12 @@ doc_fetch(int argc, VALUE *argv, VALUE self) {
     return val;
 }
 
-/* call-seq: each_leaf(path=nil) => nil
+/* @overload each_leaf(path=nil) => nil
  *
  * Yields to the provided block for each leaf node with the identified
  * location of the JSON document as the root. The parameter passed to the
  * block on yield is the Doc instance after moving to the child location.
- * @param [String] path if provided it identified the top of the branch to process the leaves of
+ *   @param [String] path if provided it identified the top of the branch to process the leaves of
  * @yieldparam [Doc] Doc at the child location
  * @example
  *   Oj::Doc.open('[3,[2,1]]') { |doc|
@@ -1429,11 +1474,11 @@ doc_each_leaf(int argc, VALUE *argv, VALUE self) {
     return Qnil;
 }
 
-/* call-seq: move(path) => nil
+/* @overload move(path) => nil
  *
  * Moves the document marker to the path specified. The path can an absolute
  * path or a relative path.
- * @param [String] path path to the location to move to
+ *   @param [String] path path to the location to move to
  * @example
  *   Oj::Doc.open('{"one":[1,2]') { |doc| doc.move('/one/2'); doc.where? }  #=> "/one/2"
  */
@@ -1455,13 +1500,13 @@ doc_move(VALUE self, VALUE str) {
     return Qnil;
 }
 
-/* call-seq: each_child(path=nil) { |doc| ... } => nil
+/* @overload each_child(path=nil) { |doc| ... } => nil
  *
  * Yields to the provided block for each immediate child node with the
  * identified location of the JSON document as the root. The parameter passed
  * to the block on yield is the Doc instance after moving to the child
  * location.
- * @param [String] path if provided it identified the top of the branch to process the chilren of
+ *   @param [String] path if provided it identified the top of the branch to process the chilren of
  * @yieldparam [Doc] Doc at the child location
  * @example
  *   Oj::Doc.open('[3,[2,1]]') { |doc|
@@ -1515,13 +1560,13 @@ doc_each_child(int argc, VALUE *argv, VALUE self) {
     return Qnil;
 }
 
-/* call-seq: each_value(path=nil) { |val| ... } => nil
+/* @overload each_value(path=nil) { |val| ... } => nil
  *
  * Yields to the provided block for each leaf value in the identified location
  * of the JSON document. The parameter passed to the block on yield is the
  * value of the leaf. Only those leaves below the element specified by the
  * path parameter are processed.
- * @param [String] path if provided it identified the top of the branch to process the leaf values of
+ *   @param [String] path if provided it identified the top of the branch to process the leaf values of
  * @yieldparam [Object] val each leaf value
  * @example
  *   Oj::Doc.open('[3,[2,1]]') { |doc|
@@ -1556,12 +1601,12 @@ doc_each_value(int argc, VALUE *argv, VALUE self) {
     return Qnil;
 }
 
-/* call-seq: dump(path=nil) => String
+/* @overload dump(path, filename)
  *
  * Dumps the document or nodes to a new JSON document. It uses the default
  * options for generating the JSON.
- * @param [String] path if provided it identified the top of the branch to dump to JSON
- * @param [String] filename if provided it is the filename to write the output to
+ *   @param path [String] if provided it identified the top of the branch to dump to JSON
+ *   @param filename [String] if provided it is the filename to write the output to
  * @example
  *   Oj::Doc.open('[3,[2,1]]') { |doc|
  *       doc.dump('/2')
@@ -1586,15 +1631,15 @@ doc_dump(int argc, VALUE *argv, VALUE self) {
 	}
     }
     if (0 != (leaf = get_doc_leaf(doc, path))) {
-	VALUE	rjson;
+	volatile VALUE	rjson;
 
 	if (0 == filename) {
 	    char	buf[4096];
-	    struct _Out out;
+	    struct _out out;
 
 	    out.buf = buf;
 	    out.end = buf + sizeof(buf) - 10;
-	    out.allocated = 0;
+	    out.allocated = false;
 	    out.omit_nil = oj_default_options.dump_opts.omit_nil;
 	    oj_dump_leaf_to_json(leaf, &oj_default_options, &out);
 	    rjson = rb_str_new2(out.buf);
@@ -1610,7 +1655,7 @@ doc_dump(int argc, VALUE *argv, VALUE self) {
     return Qnil;
 }
 
-/* call-seq: size() => Fixnum
+/* @overload size() => Fixnum
  *
  * Returns the number of nodes in the JSON document where a node is any one of
  * the basic JSON components.
@@ -1623,7 +1668,7 @@ doc_size(VALUE self) {
     return ULONG2NUM(((Doc)DATA_PTR(self))->size);
 }
 
-/* call-seq: close() => nil
+/* @overload close() => nil
  *
  * Closes an open document. No further calls to the document will be valid
  * after closing.
