@@ -8,7 +8,7 @@
 #include "dump.h"
 #include "parse.h"
 
-static VALUE	symbolize_names_sym;
+static VALUE	symbolize_names_sym = Qundef;
 
 extern const char	oj_json_class[];
 
@@ -26,31 +26,31 @@ static VALUE	state_class;
 // mimic JSON documentation
 
 /* Document-module: JSON::Ext
- * 
+ *
  * The Ext module is a placeholder in the mimic JSON module used for
  * compatibility only.
  */
 /* Document-class: JSON::Ext::Parser
- * 
+ *
  * The JSON::Ext::Parser is a placeholder in the mimic JSON module used for
  * compatibility only.
  */
 /* Document-class: JSON::Ext::Generator
- * 
+ *
  * The JSON::Ext::Generator is a placeholder in the mimic JSON module used for
  * compatibility only.
  */
 
 /* Document-method: parser=
  * call-seq: parser=(parser)
- * 
- * Does nothing other than provide compatibiltiy.
+ *
+ * Does nothing other than provide compatibility.
  * - *parser* [_Object_] ignored
  */
 /* Document-method: generator=
  * call-seq: generator=(generator)
- * 
- * Does nothing other than provide compatibiltiy.
+ *
+ * Does nothing other than provide compatibility.
  * - *generator* [_Object_] ignored
  */
 
@@ -167,7 +167,7 @@ oj_parse_mimic_dump_options(VALUE ropts, Options copts) {
     if (Qnil != (v = rb_hash_lookup(ropts, oj_ascii_only_sym))) {
 	// generate seems to assume anything except nil and false are true.
 	if (Qfalse == v) {
-	    copts->escape_mode = JXEsc; // JSONEsc;
+	    copts->escape_mode = JXEsc;
 	} else {
 	    copts->escape_mode = ASCIIEsc;
 	}
@@ -184,9 +184,9 @@ mimic_limit_arg(VALUE a) {
 
 /* Document-method: dump
  * call-seq: dump(obj, anIO=nil, limit=nil)
- * 
+ *
  * Encodes an object as a JSON String.
- * 
+ *
  * - *obj* [_Object_] object to convert to encode as JSON
  * - *anIO* [_IO_] an IO that allows writing
  * - *limit* [_Fixnum_] ignored
@@ -196,8 +196,8 @@ mimic_limit_arg(VALUE a) {
 static VALUE
 mimic_dump(int argc, VALUE *argv, VALUE self) {
     char		buf[4096];
-    struct _Out		out;
-    struct _Options	copts = oj_default_options;
+    struct _out		out;
+    struct _options	copts = oj_default_options;
     VALUE		rstr;
 
     copts.str_rx.head = NULL;
@@ -218,7 +218,7 @@ mimic_dump(int argc, VALUE *argv, VALUE self) {
     out.omit_nil = copts.dump_opts.omit_nil;
     if (2 <= argc) {
 	int	limit;
-	
+
 	// The json gem take a more liberal approach to optional
 	// arguments. Expected are (obj, anIO=nil, limit=nil) yet the io
 	// argument can be left off completely and the 2nd argument is then
@@ -275,25 +275,21 @@ mimic_walk(VALUE key, VALUE obj, VALUE proc) {
 	    rb_yield(obj);
 	}
     } else {
-#if HAS_PROC_WITH_BLOCK
 	VALUE	args[1];
 
 	*args = obj;
 	rb_proc_call_with_block(proc, 1, args, Qnil);
-#else
-	rb_raise(rb_eNotImpError, "Calling a Proc with a block not supported in this version. Use func() {|x| } syntax instead.");
-#endif
     }
     return ST_CONTINUE;
 }
 
 /* Document-method: restore
  * call-seq: restore(source, proc=nil)
- * 
+ *
  * Loads a Ruby Object from a JSON source that can be either a String or an
  * IO. If Proc is given or a block is providedit is called with each nested
  * element of the loaded Object.
- * 
+ *
  * - *source* [_String_|IO] JSON source
  * - *proc* [_Proc_] to yield to on each element or nil
  *
@@ -302,11 +298,11 @@ mimic_walk(VALUE key, VALUE obj, VALUE proc) {
 
 /* Document-method: load
  * call-seq: load(source, proc=nil)
- * 
+ *
  * Loads a Ruby Object from a JSON source that can be either a String or an
  * IO. If Proc is given or a block is providedit is called with each nested
  * element of the loaded Object.
- * 
+ *
  * - *source* [_String_|IO] JSON source
  * - *proc* [_Proc_] to yield to on each element or nil
  *
@@ -317,7 +313,7 @@ mimic_load(int argc, VALUE *argv, VALUE self) {
     VALUE	obj;
     VALUE	p = Qnil;
 
-    obj = oj_compat_parse(argc, argv, self);
+    obj = oj_compat_load(argc, argv, self);
     if (2 <= argc) {
 	if (rb_cProc == rb_obj_class(argv[1])) {
 	    p = argv[1];
@@ -334,10 +330,10 @@ mimic_load(int argc, VALUE *argv, VALUE self) {
 
 /* Document-method: []
  * call-seq: [](obj, opts={})
- * 
+ *
  * If the obj argument is a String then it is assumed to be a JSON String and
  * parsed otherwise the obj is encoded as a JSON String.
- * 
+ *
  * - *obj* [_String_|Hash|Array] object to convert
  * - *opts* [_Hash_] same options as either generate or parse
  *
@@ -358,7 +354,7 @@ mimic_dump_load(int argc, VALUE *argv, VALUE self) {
 static VALUE
 mimic_generate_core(int argc, VALUE *argv, Options copts) {
     char	buf[4096];
-    struct _Out	out;
+    struct _out	out;
     VALUE	rstr;
 
     out.buf = buf;
@@ -370,6 +366,7 @@ mimic_generate_core(int argc, VALUE *argv, Options copts) {
     // it is.
     copts->dump_opts.nan_dump = RaiseNan;
     copts->mode = CompatMode;
+    copts->to_json = Yes;
     if (2 == argc && Qnil != argv[1]) {
 	oj_parse_mimic_dump_options(argv[1], copts);
     }
@@ -399,11 +396,11 @@ mimic_generate_core(int argc, VALUE *argv, Options copts) {
 
 /* Document-method: generate
  * call-seq: generate(obj, opts=nil)
- * 
+ *
  * Encode obj as a JSON String. The obj argument must be a Hash, Array, or
  * respond to to_h or to_json. Options other than those listed such as
  * +:allow_nan+ or +:max_nesting+ are ignored.
- * 
+ *
  * - *obj* [_Object_|Hash|Array] object to convert to a JSON String
  * - *opts* [_Hash_] options
  * - - *:indent* [_String_] String to use for indentation.
@@ -417,7 +414,7 @@ mimic_generate_core(int argc, VALUE *argv, Options copts) {
  */
 VALUE
 oj_mimic_generate(int argc, VALUE *argv, VALUE self) {
-    struct _Options	copts = oj_default_options;
+    struct _options	copts = oj_default_options;
 
     copts.str_rx.head = NULL;
     copts.str_rx.tail = NULL;
@@ -435,7 +432,7 @@ oj_mimic_generate(int argc, VALUE *argv, VALUE self) {
  */
 VALUE
 oj_mimic_pretty_generate(int argc, VALUE *argv, VALUE self) {
-    struct _Options	copts = oj_default_options;
+    struct _options	copts = oj_default_options;
     VALUE		rargs[2];
     volatile VALUE	h;
 
@@ -464,7 +461,7 @@ oj_mimic_pretty_generate(int argc, VALUE *argv, VALUE self) {
 	rb_hash_aset(h, oj_array_nl_sym, rb_str_new2("\n"));
     }
     rargs[1] = rb_funcall(state_class, oj_new_id, 1, h);
-    
+
     copts.str_rx.head = NULL;
     copts.str_rx.tail = NULL;
     strcpy(copts.dump_opts.indent_str, "  ");
@@ -484,15 +481,14 @@ oj_mimic_pretty_generate(int argc, VALUE *argv, VALUE self) {
 
 static VALUE
 mimic_parse_core(int argc, VALUE *argv, VALUE self, bool bang) {
-    struct _ParseInfo	pi;
+    struct _parseInfo	pi;
+    VALUE		ropts;
     VALUE		args[1];
 
-    if (argc < 1) {
-	rb_raise(rb_eArgError, "Wrong number of arguments to parse.");
-    }
+    rb_scan_args(argc, argv, "11", NULL, &ropts);
     parse_info_init(&pi);
     oj_set_compat_callbacks(&pi);
-    // TBD
+
     pi.err_class = oj_json_parser_error_class;
     //pi.err_class = Qnil;
 
@@ -508,13 +504,15 @@ mimic_parse_core(int argc, VALUE *argv, VALUE self, bool bang) {
     pi.options.mode = CompatMode;
     pi.max_depth = 100;
 
-    if (2 <= argc && Qnil != argv[1]) {
-	VALUE	ropts = argv[1];
+    if (Qnil != ropts) {
 	VALUE	v;
 
 	if (T_HASH != rb_type(ropts)) {
 	    rb_raise(rb_eArgError, "options must be a hash.");
 	}
+	if (Qundef == symbolize_names_sym) {
+	    symbolize_names_sym = ID2SYM(rb_intern("symbolize_names"));	rb_gc_register_address(&symbolize_names_sym);
+	}
 	if (Qnil != (v = rb_hash_lookup(ropts, symbolize_names_sym))) {
 	    pi.options.sym_key = (Qtrue == v) ? Yes : No;
 	}
@@ -607,9 +605,9 @@ mimic_parse_bang(int argc, VALUE *argv, VALUE self) {
 
 /* Document-method: recurse_proc
  * call-seq: recurse_proc(obj, &proc)
- * 
- * Yields to the proc for every element in the obj recursivly.
- * 
+ *
+ * Yields to the proc for every element in the obj recursively.
+ *
  * - *obj* [_Hash_|Array] object to walk
  * - *proc* [_Proc_] to yield to on each element
  */
@@ -636,7 +634,7 @@ mimic_set_create_id(VALUE self, VALUE id) {
     Check_Type(id, T_STRING);
 
     if (NULL != oj_default_options.create_id) {
-	if (oj_json_class != oj_default_options.create_id && NULL != oj_default_options.create_id) {
+	if (oj_json_class != oj_default_options.create_id) {
 	    xfree((char*)oj_default_options.create_id);
 	}
 	oj_default_options.create_id = NULL;
@@ -665,7 +663,7 @@ mimic_create_id(VALUE self) {
     return rb_str_new_cstr(oj_json_class);
 }
 
-static struct _Options	mimic_object_to_json_options = {
+static struct _options	mimic_object_to_json_options = {
     0,		// indent
     No,		// circular
     No,		// auto_define
@@ -680,12 +678,15 @@ static struct _Options	mimic_object_to_json_options = {
     No,		// to_json
     No,		// as_json
     No,		// nilnil
-    Yes,	// empty_string
+    No,		// empty_string
     Yes,	// allow_gc
     Yes,	// quirks_mode
     No,		// allow_invalid
     No,		// create_ok
     No,		// allow_nan
+    No,		// trace
+    0,		// integer_range_min
+    0,		// integer_range_max
     oj_json_class,// create_id
     10,		// create_id_len
     3,		// sec_prec
@@ -719,9 +720,9 @@ static struct _Options	mimic_object_to_json_options = {
 static VALUE
 mimic_object_to_json(int argc, VALUE *argv, VALUE self) {
     char		buf[4096];
-    struct _Out		out;
+    struct _out		out;
     VALUE		rstr;
-    struct _Options	copts = oj_default_options;
+    struct _options	copts = oj_default_options;
 
     copts.str_rx.head = NULL;
     copts.str_rx.tail = NULL;
@@ -836,7 +837,7 @@ oj_define_mimic_json(int argc, VALUE *argv, VALUE self) {
     VALUE	dummy;
     VALUE	verbose;
     VALUE	json;
-    
+
     // Either set the paths to indicate JSON has been loaded or replaces the
     // methods if it has been loaded.
     if (rb_const_defined_at(rb_cObject, rb_intern("JSON"))) {
@@ -859,9 +860,6 @@ oj_define_mimic_json(int argc, VALUE *argv, VALUE self) {
 	    rb_funcall2(Oj, rb_intern("mimic_loaded"), 0, 0);
 	}
     }
-
-    // TBD create all modules in mimic_loaded
-
     oj_mimic_json_methods(json);
 
     rb_define_method(rb_cObject, "to_json", mimic_object_to_json, -1);
