@@ -3,6 +3,7 @@
  * All rights reserved.
  */
 
+#include <stdint.h>
 #include <stdio.h>
 #include <time.h>
 
@@ -13,6 +14,8 @@
 #include "hash.h"
 #include "odd.h"
 #include "encode.h"
+#include "trace.h"
+#include "util.h"
 
 inline static long
 read_long(const char *str, size_t len) {
@@ -286,28 +289,23 @@ hat_num(ParseInfo pi, Val parent, Val kval, NumInfo ni) {
 		if (86400 == ni->exp) { // UTC time
 		    parent->val = rb_time_nano_new(ni->i, (long)nsec);
 		    // Since the ruby C routines alway create local time, the
-		    // offset and then a convertion to UTC keeps makes the time
+		    // offset and then a conversion to UTC keeps makes the time
 		    // match the expected value.
 		    parent->val = rb_funcall2(parent->val, oj_utc_id, 0, 0);
 		} else if (ni->hasExp) {
-		    time_t	t = (time_t)(ni->i + ni->exp);
-		    struct tm	*st = gmtime(&t);
-		    VALUE	args[8];
+		    int64_t		t = (int64_t)(ni->i + ni->exp);
+		    struct _timeInfo	ti;
+		    VALUE		args[8];
 
-		    // Windows does not support dates before 1970 so ignore
-		    // the zone and do the best we can.
-		    if (NULL == st) {
-			parent->val = rb_time_nano_new(ni->i, (long)nsec);
-		    } else {
-			args[0] = LONG2NUM((long)(1900 + st->tm_year));
-			args[1] = LONG2NUM(1 + st->tm_mon);
-			args[2] = LONG2NUM(st->tm_mday);
-			args[3] = LONG2NUM(st->tm_hour);
-			args[4] = LONG2NUM(st->tm_min);
-			args[5] = rb_float_new((double)st->tm_sec + ((double)nsec + 0.5) / 1000000000.0);
-			args[6] = LONG2NUM(ni->exp);
-			parent->val = rb_funcall2(rb_cTime, oj_new_id, 7, args);
-		    }
+		    sec_as_time(t, &ti);
+		    args[0] = LONG2NUM((long)(ti.year));
+		    args[1] = LONG2NUM(ti.mon);
+		    args[2] = LONG2NUM(ti.day);
+		    args[3] = LONG2NUM(ti.hour);
+		    args[4] = LONG2NUM(ti.min);
+		    args[5] = rb_float_new((double)ti.sec + ((double)nsec + 0.5) / 1000000000.0);
+		    args[6] = LONG2NUM(ni->exp);
+		    parent->val = rb_funcall2(rb_cTime, oj_new_id, 7, args);
 		} else {
 		    parent->val = rb_time_nano_new(ni->i, (long)nsec);
 		}
@@ -341,7 +339,7 @@ hat_value(ParseInfo pi, Val parent, const char *key, size_t klen, volatile VALUE
             volatile VALUE	sc;
 	    volatile VALUE	e1;
 	    int			slen;
-	    
+
 	    if (0 == len) {
 		oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "Invalid struct data");
 		return 1;
@@ -366,11 +364,11 @@ hat_value(ParseInfo pi, Val parent, const char *key, size_t klen, volatile VALUE
             parent->val = rb_obj_alloc(sc);
             // If the JSON array has more entries than the struct class allows, we record an error.
 #ifdef RSTRUCT_LEN
-#if UNIFY_FIXNUM_AND_BIGNUM
+#if RSTRUCT_LEN_RETURNS_INTEGER_OBJECT
 	    slen = (int)NUM2LONG(RSTRUCT_LEN(parent->val));
-#else // UNIFY_FIXNUM_AND_INTEGER
+#else // RSTRUCT_LEN_RETURNS_INTEGER_OBJECT
 	    slen = (int)RSTRUCT_LEN(parent->val);
-#endif // UNIFY_FIXNUM_AND_INTEGER
+#endif // RSTRUCT_LEN_RETURNS_INTEGER_OBJECT
 #else
 	    slen = FIX2INT(rb_funcall2(parent->val, oj_length_id, 0, 0));
 #endif
@@ -380,14 +378,14 @@ hat_value(ParseInfo pi, Val parent, const char *key, size_t klen, volatile VALUE
             } else {
 		int	i;
 
-		for (i = 0; i < slen; i++) {
+		for (i = 0; i < len - 1; i++) {
 		    rb_struct_aset(parent->val, INT2FIX(i), RARRAY_PTR(value)[i + 1]);
 		}
             }
 	    return 1;
 	} else if (3 <= klen && '#' == key[1]) {
 	    volatile VALUE	*a;
-	
+
 	    if (2 != len) {
 		oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "invalid hash pair");
 		return 1;
@@ -402,23 +400,6 @@ hat_value(ParseInfo pi, Val parent, const char *key, size_t klen, volatile VALUE
     return 0;
 }
 
-static void
-copy_ivars(VALUE target, VALUE src) {
-    volatile VALUE	vars = rb_funcall(src, oj_instance_variables_id, 0);
-    volatile VALUE	*np = RARRAY_PTR(vars);
-    ID			vid;
-    int			i, cnt = (int)RARRAY_LEN(vars);
-    const char		*attr;
-
-    for (i = cnt; 0 < i; i--, np++) {
-	vid = rb_to_id(*np);
-	attr = rb_id2name(vid);
-	if ('@' == *attr) {
-	    rb_ivar_set(target, vid, rb_ivar_get(src, vid));
-	}
-    }
-}
-
 void
 oj_set_obj_ivar(Val parent, Val kval, VALUE value) {
     const char	*key = kval->key;
@@ -426,21 +407,9 @@ oj_set_obj_ivar(Val parent, Val kval, VALUE value) {
     ID		var_id;
     ID		*slot;
 
-    if ('~' == *key && Qtrue == rb_obj_is_kind_of(parent->val, rb_eException)) {
-	if (5 == klen && 0 == strncmp("~mesg", key, klen)) {
-	    VALUE		args[1];
-	    volatile VALUE	prev = parent->val;
-
-	    args[0] = value;
-	    parent->val = rb_class_new_instance(1, args, rb_class_of(parent->val));
-	    copy_ivars(parent->val, prev);
-	} else if (3 == klen && 0 == strncmp("~bt", key, klen)) {
-	    rb_funcall(parent->val, rb_intern("set_backtrace"), 1, value);
-	}
-    }
-#if USE_PTHREAD_MUTEX
+#if HAVE_LIBPTHREAD
     pthread_mutex_lock(&oj_cache_mutex);
-#elif USE_RB_MUTEX
+#else
     rb_mutex_lock(oj_cache_mutex);
 #endif
     if (0 == (var_id = oj_attr_hash_get(key, klen, &slot))) {
@@ -472,9 +441,9 @@ oj_set_obj_ivar(Val parent, Val kval, VALUE value) {
 	}
 	*slot = var_id;
     }
-#if USE_PTHREAD_MUTEX
+#if HAVE_LIBPTHREAD
     pthread_mutex_unlock(&oj_cache_mutex);
-#elif USE_RB_MUTEX
+#else
     rb_mutex_unlock(oj_cache_mutex);
 #endif
     rb_ivar_set(parent->val, var_id, value);
@@ -482,14 +451,15 @@ oj_set_obj_ivar(Val parent, Val kval, VALUE value) {
 
 static void
 hash_set_cstr(ParseInfo pi, Val kval, const char *str, size_t len, const char *orig) {
-    const char	*key = kval->key;
-    int		klen = kval->klen;
-    Val		parent = stack_peek(&pi->stack);
+    const char		*key = kval->key;
+    int			klen = kval->klen;
+    Val			parent = stack_peek(&pi->stack);
+    volatile VALUE	rval = Qnil;
 
  WHICH_TYPE:
     switch (rb_type(parent->val)) {
     case T_NIL:
-	parent->odd_args = 0; // make sure it is 0 in case not odd
+	parent->odd_args = NULL; // make sure it is NULL in case not odd
 	if ('^' != *key || !hat_cstr(pi, parent, kval, str, len)) {
 	    parent->val = rb_hash_new();
 	    goto WHICH_TYPE;
@@ -499,81 +469,98 @@ hash_set_cstr(ParseInfo pi, Val kval, const char *str, size_t len, const char *o
 	rb_hash_aset(parent->val, calc_hash_key(pi, kval, parent->k1), str_to_value(pi, str, len, orig));
 	break;
     case T_STRING:
+	rval = str_to_value(pi, str, len, orig);
 	if (4 == klen && 's' == *key && 'e' == key[1] && 'l' == key[2] && 'f' == key[3]) {
-	    rb_funcall(parent->val, oj_replace_id, 1, str_to_value(pi, str, len, orig));
+	    rb_funcall(parent->val, oj_replace_id, 1, rval);
 	} else {
-	    oj_set_obj_ivar(parent, kval, str_to_value(pi, str, len, orig));
+	    oj_set_obj_ivar(parent, kval, rval);
 	}
 	break;
     case T_OBJECT:
-	oj_set_obj_ivar(parent, kval, str_to_value(pi, str, len, orig));
+	rval = str_to_value(pi, str, len, orig);
+	oj_set_obj_ivar(parent, kval, rval);
 	break;
     case T_CLASS:
-	if (0 == parent->odd_args) {
+	if (NULL == parent->odd_args) {
 	    oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "%s is not an odd class", rb_class2name(rb_obj_class(parent->val)));
 	    return;
-	} else if (0 != oj_odd_set_arg(parent->odd_args, kval->key, kval->klen, str_to_value(pi, str, len, orig))) {
-	    char	buf[256];
+	} else {
+	    rval = str_to_value(pi, str, len, orig);
+	    if (0 != oj_odd_set_arg(parent->odd_args, kval->key, kval->klen, rval)) {
+		char	buf[256];
 
-	    if ((int)sizeof(buf) - 1 <= klen) {
-		klen = sizeof(buf) - 2;
+		if ((int)sizeof(buf) - 1 <= klen) {
+		    klen = sizeof(buf) - 2;
+		}
+		memcpy(buf, key, klen);
+		buf[klen] = '\0';
+		oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "%s is not an attribute of %s", buf, rb_class2name(rb_obj_class(parent->val)));
 	    }
-	    memcpy(buf, key, klen);
-	    buf[klen] = '\0';
-	    oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "%s is not an attribute of %s", buf, rb_class2name(rb_obj_class(parent->val)));
 	}
 	break;
     default:
 	oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "can not add attributes to a %s", rb_class2name(rb_obj_class(parent->val)));
 	return;
     }
+    if (Yes == pi->options.trace) {
+	oj_trace_parse_call("set_string", pi, __FILE__, __LINE__, rval);
+    }
 }
 
 static void
 hash_set_num(ParseInfo pi, Val kval, NumInfo ni) {
-    const char	*key = kval->key;
-    int		klen = kval->klen;
-    Val		parent = stack_peek(&pi->stack);
+    const char		*key = kval->key;
+    int			klen = kval->klen;
+    Val			parent = stack_peek(&pi->stack);
+    volatile VALUE	rval = Qnil;
 
  WHICH_TYPE:
     switch (rb_type(parent->val)) {
     case T_NIL:
-	parent->odd_args = 0; // make sure it is 0 in case not odd
+	parent->odd_args = NULL; // make sure it is NULL in case not odd
 	if ('^' != *key || !hat_num(pi, parent, kval, ni)) {
 	    parent->val = rb_hash_new();
 	    goto WHICH_TYPE;
 	}
 	break;
     case T_HASH:
-	rb_hash_aset(parent->val, calc_hash_key(pi, kval, parent->k1), oj_num_as_value(ni));
+	rval = oj_num_as_value(ni);
+	rb_hash_aset(parent->val, calc_hash_key(pi, kval, parent->k1), rval);
 	break;
     case T_OBJECT:
 	if (2 == klen && '^' == *key && 'i' == key[1] &&
 	    !ni->infinity && !ni->neg && 1 == ni->div && 0 == ni->exp && 0 != pi->circ_array) { // fixnum
 	    oj_circ_array_set(pi->circ_array, parent->val, ni->i);
 	} else {
-	    oj_set_obj_ivar(parent, kval, oj_num_as_value(ni));
+	    rval = oj_num_as_value(ni);
+	    oj_set_obj_ivar(parent, kval, rval);
 	}
 	break;
     case T_CLASS:
-	if (0 == parent->odd_args) {
+	if (NULL == parent->odd_args) {
 	    oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "%s is not an odd class", rb_class2name(rb_obj_class(parent->val)));
 	    return;
-	} else if (0 != oj_odd_set_arg(parent->odd_args, key, klen, oj_num_as_value(ni))) {
-	    char	buf[256];
+	} else {
+	    rval = oj_num_as_value(ni);
+	    if (0 != oj_odd_set_arg(parent->odd_args, key, klen, rval)) {
+		char	buf[256];
 
-	    if ((int)sizeof(buf) - 1 <= klen) {
-		klen = sizeof(buf) - 2;
+		if ((int)sizeof(buf) - 1 <= klen) {
+		    klen = sizeof(buf) - 2;
+		}
+		memcpy(buf, key, klen);
+		buf[klen] = '\0';
+		oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "%s is not an attribute of %s", buf, rb_class2name(rb_obj_class(parent->val)));
 	    }
-	    memcpy(buf, key, klen);
-	    buf[klen] = '\0';
-	    oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "%s is not an attribute of %s", buf, rb_class2name(rb_obj_class(parent->val)));
 	}
 	break;
     default:
 	oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "can not add attributes to a %s", rb_class2name(rb_obj_class(parent->val)));
 	return;
     }
+    if (Yes == pi->options.trace) {
+	oj_trace_parse_call("add_number", pi, __FILE__, __LINE__, rval);
+    }
 }
 
 static void
@@ -585,7 +572,7 @@ hash_set_value(ParseInfo pi, Val kval, VALUE value) {
  WHICH_TYPE:
     switch (rb_type(parent->val)) {
     case T_NIL:
-	parent->odd_args = 0; // make sure it is 0 in case not odd
+	parent->odd_args = NULL; // make sure it is NULL in case not odd
 	if ('^' != *key || !hat_value(pi, parent, key, klen, value)) {
 	    parent->val = rb_hash_new();
 	    goto WHICH_TYPE;
@@ -602,7 +589,7 @@ hash_set_value(ParseInfo pi, Val kval, VALUE value) {
 	    if (3 <= klen && '^' == *key && '#' == key[1] && T_ARRAY == rb_type(value)) {
 		long		len = RARRAY_LEN(value);
 		volatile VALUE	*a = RARRAY_PTR(value);
-	
+
 		if (2 != len) {
 		    oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "invalid hash pair");
 		    return;
@@ -626,7 +613,7 @@ hash_set_value(ParseInfo pi, Val kval, VALUE value) {
 	break;
     case T_MODULE:
     case T_CLASS:
-	if (0 == parent->odd_args) {
+	if (NULL == parent->odd_args) {
 	    oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "%s is not an odd class", rb_class2name(rb_obj_class(parent->val)));
 	    return;
 	} else if (0 !=	oj_odd_set_arg(parent->odd_args, key, klen, value)) {
@@ -644,30 +631,41 @@ hash_set_value(ParseInfo pi, Val kval, VALUE value) {
 	oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "can not add attributes to a %s", rb_class2name(rb_obj_class(parent->val)));
 	return;
     }
+    if (Yes == pi->options.trace) {
+	oj_trace_parse_call("add_value", pi, __FILE__, __LINE__, value);
+    }
 }
 
 static VALUE
 start_hash(ParseInfo pi) {
+    if (Yes == pi->options.trace) {
+	oj_trace_parse_in("start_hash", pi, __FILE__, __LINE__);
+    }
     return Qnil;
 }
 
 static void
-end_hash(struct _ParseInfo *pi) {
+end_hash(ParseInfo pi) {
     Val	parent = stack_peek(&pi->stack);
 
     if (Qnil == parent->val) {
 	parent->val = rb_hash_new();
-    } else if (0 != parent->odd_args) {
+    } else if (NULL != parent->odd_args) {
 	OddArgs	oa = parent->odd_args;
 
 	parent->val = rb_funcall2(oa->odd->create_obj, oa->odd->create_op, oa->odd->attr_cnt, oa->args);
 	oj_odd_free(oa);
-	parent->odd_args = 0;
+	parent->odd_args = NULL;
+    }
+    if (Yes == pi->options.trace) {
+	oj_trace_parse_hash_end(pi, __FILE__, __LINE__);
     }
 }
 
 static void
 array_append_cstr(ParseInfo pi, const char *str, size_t len, const char *orig) {
+    volatile VALUE	rval = Qnil;
+    
     if (3 <= len && 0 != pi->circ_array) {
 	if ('i' == str[1]) {
 	    long	i = read_long(str + 2, len - 2);
@@ -683,25 +681,40 @@ array_append_cstr(ParseInfo pi, const char *str, size_t len, const char *orig) {
 		rb_ary_push(stack_peek(&pi->stack)->val, oj_circ_array_get(pi->circ_array, i));
 		return;
 	    }
-	    
+
 	}
     }
-    rb_ary_push(stack_peek(&pi->stack)->val, str_to_value(pi, str, len, orig));
+    rval = str_to_value(pi, str, len, orig);
+    rb_ary_push(stack_peek(&pi->stack)->val, rval);
+    if (Yes == pi->options.trace) {
+	oj_trace_parse_call("append_string", pi, __FILE__, __LINE__, rval);
+    }
 }
 
 static void
 array_append_num(ParseInfo pi, NumInfo ni) {
-    rb_ary_push(stack_peek(&pi->stack)->val, oj_num_as_value(ni));
+    volatile VALUE	rval = oj_num_as_value(ni);
+    
+    rb_ary_push(stack_peek(&pi->stack)->val, rval);
+    if (Yes == pi->options.trace) {
+	oj_trace_parse_call("append_number", pi, __FILE__, __LINE__, rval);
+    }
 }
 
 static void
 add_cstr(ParseInfo pi, const char *str, size_t len, const char *orig) {
     pi->stack.head->val = str_to_value(pi, str, len, orig);
+    if (Yes == pi->options.trace) {
+	oj_trace_parse_call("add_string", pi, __FILE__, __LINE__, pi->stack.head->val);
+    }
 }
 
 static void
 add_num(ParseInfo pi, NumInfo ni) {
     pi->stack.head->val = oj_num_as_value(ni);
+    if (Yes == pi->options.trace) {
+	oj_trace_parse_call("add_num", pi, __FILE__, __LINE__, pi->stack.head->val);
+    }
 }
 
 void
@@ -720,7 +733,7 @@ oj_set_object_callbacks(ParseInfo pi) {
 
 VALUE
 oj_object_parse(int argc, VALUE *argv, VALUE self) {
-    struct _ParseInfo	pi;
+    struct _parseInfo	pi;
 
     parse_info_init(&pi);
     pi.options = oj_default_options;
@@ -737,7 +750,7 @@ oj_object_parse(int argc, VALUE *argv, VALUE self) {
 
 VALUE
 oj_object_parse_cstr(int argc, VALUE *argv, char *json, size_t len) {
-    struct _ParseInfo	pi;
+    struct _parseInfo	pi;
 
     parse_info_init(&pi);
     pi.options = oj_default_options;
