@@ -19,7 +19,7 @@
 #include "oj.h"
 #include "encode.h"
 
-typedef struct _ParseInfo {
+typedef struct _parseInfo {
     char	*str;		/* buffer being read from */
     char	*s;		/* current position in buffer */
     void	*stack_min;
@@ -52,7 +52,7 @@ static void	skip_comment(ParseInfo pi);
  * a SAX parser because it uses callback when document elements are
  * encountered.
  *
- * Parsing is very tolerant. Lack of headers and even mispelled element
+ * Parsing is very tolerant. Lack of headers and even misspelled element
  * endings are passed over without raising an error. A best attempt is made in
  * all cases to parse the string.
  */
@@ -381,7 +381,7 @@ read_num(ParseInfo pi, const char *key) {
 	
 	    *pi->s = '\0';
 	    if (pi->has_add_value) {
-		call_add_value(pi->handler, rb_funcall(oj_bigdecimal_class, oj_new_id, 1, rb_str_new2(start)), key);
+		call_add_value(pi->handler, rb_funcall(rb_cObject, oj_bigdecimal_id, 1, rb_str_new2(start)), key);
 	    }
 	    *pi->s = c;
 	} else {
@@ -399,7 +399,7 @@ read_num(ParseInfo pi, const char *key) {
 	
 	    *pi->s = '\0';
 	    if (pi->has_add_value) {
-		call_add_value(pi->handler, rb_funcall(oj_bigdecimal_class, oj_new_id, 1, rb_str_new2(start)), key);
+		call_add_value(pi->handler, rb_funcall(rb_cObject, oj_bigdecimal_id, 1, rb_str_new2(start)), key);
 	    }
 	    *pi->s = c;
 	} else {
@@ -605,7 +605,7 @@ read_quoted_value(ParseInfo pi) {
 static void
 saj_parse(VALUE handler, char *json) {
     volatile VALUE	obj = Qnil;
-    struct _ParseInfo	pi;
+    struct _parseInfo	pi;
 
     if (0 == json) {
 	if (pi.has_error) {
@@ -626,7 +626,7 @@ saj_parse(VALUE handler, char *json) {
     {
 	struct rlimit	lim;
 
-	if (0 == getrlimit(RLIMIT_STACK, &lim)) {
+	if (0 == getrlimit(RLIMIT_STACK, &lim) && RLIM_INFINITY != lim.rlim_cur) {
 	    pi.stack_min = (void*)((char*)&obj - (lim.rlim_cur / 4 * 3)); /* let 3/4ths of the stack be used only */
 	} else {
 	    pi.stack_min = 0; /* indicates not to check stack limit */
